/*
 * Copyright 1993-2015 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

#ifndef PARTICLES_KERNEL_H
#define PARTICLES_KERNEL_H


#include "vector_types.h"
typedef unsigned int uint;

struct SimParams
{
    float3 gravity;
    float globalDamping;
    float noiseFreq;
    float noiseAmp;
    float3 cursorPos;

    float time;
    float3 noiseSpeed;
};

struct float4x4
{
    float m[16];
};

#endif
