/*
 * Copyright 1993-2015 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */



#ifndef REALTYPE_H
#define REALTYPE_H

//#define DOUBLE_PRECISION


#ifndef DOUBLE_PRECISION
typedef float real;
#else
typedef double real;
#endif



#endif
