/*
 * Copyright 1993-2015 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

extern const char *particleVS;
extern const char *particleSpherePS, *simplePS, *particlePS, *particleShadowPS;
extern const char *mblurVS, *mblurGS;
extern const char *passThruVS, *texture2DPS, *blurPS;
extern const char *floorVS, *floorPS;
