/*
 * Copyright 1993-2015 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and
 * proprietary rights in and to this software and related documentation.
 * Any use, reproduction, disclosure, or distribution of this software
 * and related documentation without an express license agreement from
 * NVIDIA Corporation is strictly prohibited.
 *
 * Please refer to the applicable NVIDIA end user license agreement (EULA)
 * associated with this source code for terms and conditions that govern
 * your use of this NVIDIA software.
 *
 */

////////////////////////////////////////////////////////////////////////////////
// This file is auto-generated, do not edit
////////////////////////////////////////////////////////////////////////////////
#ifndef __matrixMul_kernel_32_ptxdump_h__
#define __matrixMul_kernel_32_ptxdump_h__

#if defined __cplusplus
extern "C" {
#endif

    extern unsigned char matrixMul_kernel_32_ptxdump[25784];

#if defined __cplusplus
}
#endif

#endif //__matrixMul_kernel_32_ptxdump_h__
