#!/bin/bash

# Install to conda style directories
[[ -d lib64 ]] && mv lib64 lib

samples_dir="cuda-samples"

for i in `ls`; do
	[[ $i == "build_env_setup.sh" ]] && continue
	[[ $i == "conda_build.sh" ]] && continue
	[[ $i == "metadata_conda_debug.yaml" ]] && continue

	mkdir -p ${PREFIX}/${samples_dir}
	cp -r $i ${PREFIX}/${samples_dir}
done
